"""Collection of exceptions raised and/or processed by Cheroot."""

from __future__ import absolute_import, division, print_function
__metaclass__ = type

import errno
import sys


class MaxSizeExceeded(Exception):
    """Exception raised when a client sends more data then acceptable within limit.

    Depends on ``request.body.maxbytes`` config option if used within CherryPy
    """


class NoSSLError(Exception):
    """Exception raised when a client speaks HTTP to an HTTPS socket."""


class FatalSSLAlert(Exception):
    """Exception raised when the SSL implementation signals a fatal alert."""


def plat_specific_errors(*errnames):
    """Return error numbers for all errors in errnames on this platform.

    The 'errno' module contains different global constants depending on
    the specific platform (OS). This function will return the list of
    numeric values for a given list of potential names.
    """
    errno_names = dir(errno)
    nums = [getattr(errno, k) for k in errnames if k in errno_names]
    # de-dupe the list
    return list(dict.fromkeys(nums).keys())


socket_error_eintr = plat_specific_errors('EINTR', 'WSAEINTR')

socket_errors_to_ignore = plat_specific_errors(
    'EPIPE',
    'EBADF', 'WSAEBADF',
    'ENOTSOCK', 'WSAENOTSOCK',
    'ETIMEDOUT', 'WSAETIMEDOUT',
    'ECONNREFUSED', 'WSAECONNREFUSED',
    'ECONNRESET', 'WSAECONNRESET',
    'ECONNABORTED', 'WSAECONNABORTED',
    'ENETRESET', 'WSAENETRESET',
    'EHOSTDOWN', 'EHOSTUNREACH',
)
socket_errors_to_ignore.append('timed out')
socket_errors_to_ignore.append('The read operation timed out')
socket_errors_nonblocking = plat_specific_errors(
    'EAGAIN', 'EWOULDBLOCK', 'WSAEWOULDBLOCK')

if sys.platform == 'darwin':
    socket_errors_to_ignore.extend(plat_specific_errors('EPROTOTYPE'))
    socket_errors_nonblocking.extend(plat_specific_errors('EPROTOTYPE'))



# -*- coding: utf-8 -*-
"""Test suite for cross-python compatibility helpers."""

from __future__ import absolute_import, division, print_function
__metaclass__ = type

import pytest
import six

from cheroot._compat import ntob, ntou, bton


@pytest.mark.parametrize(
    'func,inp,out',
    [
        (ntob, 'bar', b'bar'),
        (ntou, 'bar', u'bar'),
        (bton, b'bar', 'bar'),
    ],
)
def test_compat_functions_positive(func, inp, out):
    """Check that compat functions work with correct input."""
    assert func(inp, encoding='utf-8') == out


@pytest.mark.parametrize(
    'func',
    [
        ntob,
        ntou,
    ],
)
def test_compat_functions_negative_nonnative(func):
    """Check that compat functions fail loudly for incorrect input."""
    non_native_test_str = b'bar' if six.PY3 else u'bar'
    with pytest.raises(TypeError):
        func(non_native_test_str, encoding='utf-8')


@pytest.mark.skip(reason='This test does not work now')
@pytest.mark.skipif(
    six.PY3,
    reason='This code path only appears in Python 2 version.',
)
def test_ntou_escape():
    """Check that ntou supports escape-encoding under Python 2."""
    expected = u''
    actual = ntou('hi'.encode('ISO-8859-1'), encoding='escape')
    assert actual == expected



"""HTTPS Tests."""
# -*- coding: utf-8 -*-
# vim: set fileencoding=utf-8 :

import os
import pytest
import ssl

from cheroot.ssl.builtin import BuiltinSSLAdapter
from cheroot.test import helper
import cheroot
import ddt


def create_wsgi_server(**conf):
    """Create test server."""
    ssl_adapter = BuiltinSSLAdapter(
        certificate=conf.pop('certificate'),
        private_key=conf.pop('private_key'),
        certificate_chain=conf.pop('certificate_chain'))
    ssl_adapter.context.verify_mode = conf.pop('verify_mode', ssl.CERT_NONE)
    server = cheroot.wsgi.Server(**conf)
    server.ssl_adapter = ssl_adapter
    return server


def ssl_file(filename):
    """Return path to given ssl file."""
    return os.path.join('cheroot/test/ssl/', filename)


class HTTPSTestBase(object):
    """Base class for HTTPS tests."""

    available_servers = {'wsgi': create_wsgi_server}
    config = {
        'certificate': 'cheroot/test/ssl/server.cert',
        'private_key': 'cheroot/test/ssl/server.key',
        'certificate_chain': 'cheroot/test/ssl/ca.cert',
    }
    script_name = '/hello'
    ssl_context = None

    @classmethod
    def setup_server(cls):
        """Set up the test server."""
        class Root(helper.Controller):

            def hello(req, resp):
                return 'Hello world!'

            handlers = {'/hello': hello}

        cls.httpserver.wsgi_app = Root()

    def setUp(self):
        """Set up."""
        super(HTTPSTestBase, self).setUp()
        self.ssl_context = ssl.create_default_context(
            cafile='cheroot/test/ssl/ca.cert')
        self.ssl_context.check_hostname = False

    def getPage(self, *args, **kw):
        """Fetch the page."""
        return super(HTTPSTestBase, self).getPage(
            self.script_name, *args,
            raise_subcls=ssl.SSLError, **kw)

    def set_client_cert(self, name):
        """Set client cert."""
        self.ssl_context.load_cert_chain(ssl_file(name + '.cert'),
                                         keyfile=ssl_file('client.key'))

    def assert_allowed(self, client_cert):
        """Assert test page can be fetched when given client cert is used."""
        self.set_client_cert(client_cert)
        self.getPage()
        self.assertStatus('200 OK')

    def assert_reject(self, client_cert):
        """Assert test page cannot be fetched.

        And an SSLError is raised when the given client cert is used.
        """
        self.set_client_cert(client_cert)
        with pytest.raises(ssl.SSLError) as context:
            self.getPage()
        return context


@ddt.ddt
class ClientCertRequiredTests(HTTPSTestBase, helper.CherootWebCase):
    """Test expected outcomes when client cert is required."""

    @classmethod
    def setup_class(cls):
        """Set up for tests."""
        cls.config.update({'verify_mode': ssl.CERT_REQUIRED})
        super(ClientCertRequiredTests, cls).setup_class()

    @ddt.data('client', 'client_ip', 'client_wildcard', 'client_wrong_host')
    def test_allow(self, client_cert):
        """Test that the given client cert is allowed to connect."""
        self.assert_allowed(client_cert)

    def test_reject_wrong_ca(self):
        """Test that the given client cert is not allowed to connect."""
        context = self.assert_reject('client_wrong_ca')
        assert 'tlsv1 alert unknown ca' in str(context.value)


@ddt.ddt
class ClientCertOptionalTests(HTTPSTestBase, helper.CherootWebCase):
    """Test expected outcomes when client cert is optional."""

    @classmethod
    def setup_class(cls):
        """Set up for tests."""
        cls.config.update({'verify_mode': ssl.CERT_OPTIONAL})
        super(ClientCertOptionalTests, cls).setup_class()

    @ddt.data('client', 'client_ip', 'client_wildcard', 'client_wrong_host')
    def test_allow(self, client_cert):
        """Test that the given client cert is allowed to connect."""
        self.assert_allowed(client_cert)

    def test_reject_wrong_ca(self):
        """Test that the given client cert is not allowed to connect."""
        context = self.assert_reject('client_wrong_ca')
        assert 'tlsv1 alert unknown ca' in str(context.value)


@ddt.ddt
class ClientCertIgnoredTests(HTTPSTestBase, helper.CherootWebCase):
    """Test expected outcomes when client cert is ignored."""

    @classmethod
    def setup_class(cls):
        """Set up for tests."""
        cls.config.update({'verify_mode': ssl.CERT_NONE})
        super(ClientCertIgnoredTests, cls).setup_class()

    @ddt.data('client', 'client_ip', 'client_wildcard', 'client_wrong_host',
              'client_wrong_ca')
    def test_allow(self, client_cert):
        """Test that the given client cert is allowed to connect."""
        self.assert_allowed(client_cert)
